import abc
from typing import Type

from .abstract import CTree

__all__ = (
    "register_rule",
    "CTreePostProc",
    "_REGISTRY",
)


class CTreePostProc(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def process(cls, ct: CTree) -> None:
        """Пост-обработка конфигурации, например изменение, добавление, удаление команд."""


_REGISTRY: dict[str, Type[CTreePostProc]] = {}


def register_rule(cls: Type[CTreePostProc]) -> Type[CTreePostProc]:
    if cls.__class__.__name__ not in _REGISTRY:
        _REGISTRY[cls.__name__] = cls

    return cls



import re

from .abstract import CTree

__all__ = (
    "AristaCT",
    "CiscoCT",
    "HuaweiCT",
)


class AristaCT(CTree):
    spaces = "   "
    undo = "no"
    section_exit = "exit"
    section_separator = "!"
    sections_require_exit = [
        r"route-map \S+ (?:deny|permit) \d+",
    ]
    sections_without_exit = []
    junk_lines = [
        r"\s*!.*",
        r"end",
    ]
    mask_patterns = [
        r".*(?:password|secret)(?: sha512)? (\S+)",
        r".*(?:key|md5)(?: 7)? (\S+)",
    ]


class CiscoCT(CTree):
    spaces = " "
    undo = "no"
    section_exit = "exit"
    section_separator = "!"
    sections_require_exit = [
        r"route-map \S+ (?:deny|permit) \d+",
    ]
    sections_without_exit = [
        r"crypto pki certificate chain \S+ / certificate(?: ca| self-signed)? \S+",
    ]
    junk_lines = [
        r"\s*!.*",
        r"Building configuration...",
        r"Current configuration : \d+ bytes",
        r"version \S+",
        r"\s*exit-address-family",
        r"end",
    ]
    mask_patterns = [
        r".*secret (?:5|9|7) (\S+)",
    ]
    new_line_mask = "<<br>>"

    @classmethod
    def _mask_banners(cls, config: str) -> str:
        banners = []
        for section in re.finditer(
            r"banner (?P<type>(?:motd|login|exec)) (?P<sep>\S+)(?P<body>.*?)(?P=sep)\n",
            config,
            re.DOTALL,
        ):
            banners.append(section.group("body"))
        for banner in banners:
            config = config.replace(banner, banner.replace("\n", cls.new_line_mask))
        return config

    @classmethod
    def _mask_certificates(cls, config: str) -> str:
        certificates = []
        for cert in re.finditer(
            r"(?<=\s)certificate(?: ca| self-signed)? \S+\n(?P<body>.*?\s+quit)(?=\n)",
            config,
            re.DOTALL,
        ):
            certificates.append(cert.group("body"))
        for certificate in certificates:
            config = config.replace(certificate, certificate.replace("\n", cls.new_line_mask))
        return config

    @classmethod
    def pre_run(cls, config: str) -> str:
        config = cls._mask_banners(config)
        config = cls._mask_certificates(config)
        return config

    def post_run(self) -> None:
        for node in self.children.values():
            if node.line.startswith(("banner motd", "banner exec", "banner login")):
                node.line = node.line.replace(self.new_line_mask, "\n")
            elif node.line.startswith("crypto pki certificate chain"):
                for certificates in node.children.values():
                    for cert_body in certificates.children.values():
                        cert_body.line = cert_body.line.replace(self.new_line_mask, "\n")
                    certificates.rebuild()
        self.rebuild()


class HuaweiCT(CTree):
    spaces = " "
    undo = "undo"
    section_exit = "quit"
    section_separator = "#"
    sections_require_exit = [
        r"route-policy \S+ (?:deny|permit) node \d+",
        r"aaa / authentication-scheme \S+",
        r"aaa / authorization-scheme \S+",
        r"aaa / accounting-scheme \S+",
        r"aaa / domain \S+",
    ]
    sections_without_exit = [
        r"xpl \S+ .*",
    ]
    junk_lines = [
        r"\s*#.*",
        r"!.*",
        r"return",
    ]
    mask_patterns = [
        r".*(?:auth-code|(?:pre-)?shared-key|password|md5|key|authentication|read) cipher (\S+)(?: \S+)*",
        r".*password irreversible-cipher (\S+)",
        r".*pass-phrase (\S+) aes",
    ]

    @classmethod
    def _remove_spaces(cls, config: str) -> str:
        # у huawei в некоторых устройствах/версиях некоторые глобальные команды
        # выглядят так, как будто они находятся внутри секции, это ломает парсинг.
        # например пробел(ы) перед ntp-service, хотя это глобальный конфиг
        # #
        #  ntp-service server disable
        #  ntp-service source-interface LoopBack0
        #  ntp-service unicast-server 1.2.3.4
        # #
        # или пробел перед http
        # #
        #  http timeout 60
        #  http secure-server ssl-policy default_policy
        #  http server enable
        # #
        # поэтому удаляем пробел из конфигурации перед анализом
        #! то, что встретилось, возможно есть еще какие-то случаи

        return re.sub(
            pattern=r"""\n\s+(
                ntp-service\s
                |np\scapwap-reassembly\s
                |set\snp\srss\s
                |clock\stimezone\s
                |http\stimeout\s
                |http\sserver\s
                |http\ssecure-server\s
                |defence\sengine\s
                |sysname\s
                |header\sshell\sinformation\s
                |snmp-agent\s?
                |info-center\s
                |ssh\s(?:server|client)
                |(?:undo )?s?telnet\s
                |ftp\s
                |ecc\speer-public-key\s
                )""",
            repl=r"\n\g<1>",
            string=config,
            flags=re.VERBOSE,
        )

    @classmethod
    def pre_run(cls, config: str) -> str:
        config = cls._remove_spaces(config)
        return config



from __future__ import annotations

import abc
import re
from pathlib import Path
from typing import Type

import yaml

from .abstract import CTree
from .factory import CTreeFactory
from .models import TaggingRule, Vendor

__all__ = (
    "CTreeParser",
    "TaggingRules",
    "TaggingRulesFile",
    "TaggingRulesDict",
)


class TaggingRules(abc.ABC):
    @property
    def rules(self) -> dict[Vendor, list[TaggingRule]]:
        """Правила для расстановки тегов в на строки конфигурации.

        Формат следующий:

        {
            "huawei": [ParsingRule-1, ParsingRule-2],
            "arista": [ParsingRule-N],
            "other_vendor": [],
        }

        ParsingRule это модель вида
        {
            "regex": "^ssl policy my_policy",
            "tags": ["ssl", "my_policy"],
        }
        """
        if not hasattr(self, "_rules"):
            self.load_rules()
        return self._rules

    @rules.setter
    def rules(self, rules: dict[Vendor, list[TaggingRule]]) -> None:
        self._rules = rules

    @abc.abstractmethod
    def load_rules(self) -> None:
        """Загрузка правил.

        Описывается конкретной реализацией класса
        """


class TaggingRulesFile(TaggingRules):
    _instances: dict[Path, TaggingRulesFile] = {}

    def __new__(cls, filename: Path | str) -> TaggingRulesFile:
        if isinstance(filename, str):
            filename = Path(Path.cwd(), filename)
        if filename not in cls._instances:
            cls._instances[filename] = super().__new__(cls)
        return cls._instances[filename]

    def __init__(self, filename: Path | str) -> None:
        if isinstance(filename, str):
            filename = Path(Path.cwd(), filename)
        self.filename = filename

    def load_rules(self) -> None:
        result = {}
        with open(self.filename, "r") as f:
            data = yaml.safe_load(f)
        for vendor, rules in (data.get("tagging-rules") or {}).items():
            if vendor not in [e.value for e in Vendor]:
                continue
            result[Vendor(vendor)] = [TaggingRule(**rule) for rule in rules]  # type: ignore[arg-type]

        self.rules = result


class TaggingRulesDict(TaggingRules):

    def __init__(self, rules_dict: dict[Vendor, list[dict[str, str | list[str]]]]) -> None:
        self.rules_dict = rules_dict

    def load_rules(self) -> None:
        result = {}
        for vendor, rules in self.rules_dict.items():
            if vendor not in [e.value for e in Vendor]:
                continue
            result[Vendor(vendor)] = [TaggingRule(**rule) for rule in rules]  # type: ignore[arg-type]

        self.rules = result


class CTreeParser:
    def __init__(self, vendor: Vendor, tagging_rules: TaggingRules | None = None) -> None:
        self._class = CTreeFactory.get_class(vendor)
        if tagging_rules is None:
            self.tagging_rules = []
        else:
            self.tagging_rules = tagging_rules.rules.get(vendor, [])

    def _get_tags(self, line: str) -> list[str] | None:
        for rule in self.tagging_rules:
            if m := re.search(rule.regex, line):
                return [*rule.tags, *m.groups()]
        return None

    def _parse(self, ct: Type[CTree], config: str) -> CTree:
        root = ct()
        section = [root]
        spaces = [0]
        previous_node: CTree = root
        for line in config.splitlines():
            if len(line.strip()) == 0:
                continue
            skip_pattern = "|".join(ct.junk_lines)  # type: ignore[arg-type]
            if re.fullmatch(skip_pattern, line):
                continue

            # число пробелов у текущей строки
            current_space = len(line) - len(line.lstrip())
            line = line.strip()

            # мы вошли в секцию
            if current_space > spaces[-1]:
                section.append(previous_node)
                spaces.append(current_space)
            # мы вышли из секции
            elif current_space < spaces[-1]:
                while current_space != spaces[-1]:
                    _ = section.pop()
                    _ = spaces.pop()

            parent = section[-1]
            if len(self.tagging_rules) != 0:
                full_line = parent._formal_path
                full_line.append(line)
                tags = self._get_tags(" / ".join(full_line))
            else:
                tags = None

            if line in parent.children:
                previous_node = parent.children[line]
            else:
                previous_node = ct(line=line, parent=parent, tags=tags)

        return root

    def parse(self, config: str) -> CTree:
        config = self._class.pre_run(config)
        root = self._parse(self._class, config)
        root.post_run()
        return root
